@@ -30,7 +30,7 @@ impl KeyedMessage {
     ///```
     ///# use eccodes::codes_handle::{ProductKind, CodesHandle, KeyType::Str};
     ///# use std::path::Path;
-    ///# use fallible_iterator::FallibleIterator;
+    ///# use eccodes::FallibleIterator;
     ///#
     ///let file_path = Path::new("./data/iceland.grib");
     ///let product_kind = ProductKind::GRIB;
@@ -191,7 +191,7 @@ impl KeyedMessage {
     ///# use eccodes::codes_handle::{ProductKind, CodesHandle, KeyedMessage, KeysIteratorFlags};
     ///# use std::path::Path;
     ///# use eccodes::codes_handle::KeyType::Str;
-    ///# use fallible_iterator::FallibleIterator;
+    ///# use eccodes::FallibleIterator;
     ///let file_path = Path::new("./data/iceland.grib");
     ///let product_kind = ProductKind::GRIB;
     ///
@@ -285,7 +285,7 @@ impl KeyedMessage {
     ///# use eccodes::codes_handle::{ProductKind, CodesHandle, KeyedMessage, KeysIteratorFlags};
     ///# use std::path::Path;
     ///# use eccodes::codes_handle::KeyType::Str;
-    ///# use fallible_iterator::FallibleIterator;
+    ///# use eccodes::FallibleIterator;
     ///let file_path = Path::new("./data/iceland.grib");
     ///let product_kind = ProductKind::GRIB;
     ///
@@ -361,7 +361,7 @@ impl Clone for KeyedMessage {
 ///# use eccodes::codes_handle::{ProductKind, CodesHandle, KeyedMessage, KeysIteratorFlags};
 ///# use std::path::Path;
 ///# use eccodes::codes_handle::KeyType::Str;
-///# use fallible_iterator::FallibleIterator;
+///# use eccodes::FallibleIterator;
 ///let file_path = Path::new("./data/iceland.grib");
 ///let product_kind = ProductKind::GRIB;
 ///
@@ -430,6 +430,8 @@ impl Drop for KeyedMessage {
             }
         }
 
+        self.nearest_handle = Some(null_mut());
+
         if let Some(kiter) = self.keys_iterator {
             unsafe {
                 codes_keys_iterator_delete(kiter).unwrap_or_else(|error| {
@@ -456,7 +458,7 @@ impl Drop for KeyedMessage {
 #[cfg(test)]
 mod tests {
     use crate::codes_handle::{CodesHandle, KeyType, KeysIteratorFlags, ProductKind};
-    use fallible_iterator::FallibleIterator;
+    use crate::FallibleIterator;
     use std::path::Path;
     use testing_logger;
 
