@@ -10,15 +10,32 @@ use std::{
     ptr,
 };
 
-use eccodes_sys::{codes_context, codes_handle, _IO_FILE};
+use eccodes_sys::{
+    codes_context, codes_handle, codes_keys_iterator, codes_nearest, CODES_NEAREST_SAME_DATA,
+    CODES_NEAREST_SAME_GRID, CODES_TYPE_BYTES, CODES_TYPE_DOUBLE, CODES_TYPE_LABEL,
+    CODES_TYPE_LONG, CODES_TYPE_MISSING, CODES_TYPE_SECTION, CODES_TYPE_STRING,
+    CODES_TYPE_UNDEFINED, _IO_FILE,
+};
 use libc::{c_void, FILE};
 use num_traits::FromPrimitive;
 
 use crate::{
-    codes_handle::ProductKind,
+    codes_handle::{NearestGridpoint, ProductKind},
     errors::{CodesError, CodesInternal},
 };
 
+#[derive(Copy, Eq, PartialEq, Clone, Ord, PartialOrd, Hash, Debug, num_derive::FromPrimitive)]
+pub enum NativeKeyType {
+    Undefined = CODES_TYPE_UNDEFINED as isize,
+    Long = CODES_TYPE_LONG as isize,
+    Double = CODES_TYPE_DOUBLE as isize,
+    Str = CODES_TYPE_STRING as isize,
+    Bytes = CODES_TYPE_BYTES as isize,
+    Section = CODES_TYPE_SECTION as isize,
+    Label = CODES_TYPE_LABEL as isize,
+    Missing = CODES_TYPE_MISSING as isize,
+}
+
 pub unsafe fn codes_handle_new_from_file(
     file_pointer: *mut FILE,
     product_kind: ProductKind,
@@ -53,13 +70,6 @@ pub unsafe fn codes_handle_delete(handle: *mut codes_handle) -> Result<(), Codes
     Ok(())
 }
 
-#[derive(Copy, Eq, PartialEq, Clone, Ord, PartialOrd, Hash, Debug, num_derive::FromPrimitive)]
-pub enum NativeKeyType {
-    Long = 1,
-    Double = 2,
-    Str = 3,
-}
-
 pub unsafe fn codes_get_native_type(
     handle: *mut codes_handle,
     key: &str,
@@ -203,11 +213,39 @@ pub unsafe fn codes_get_string(handle: *mut codes_handle, key: &str) -> Result<S
     }
 
     key_message.truncate(key_length as usize);
-    let key_message = CStr::from_bytes_with_nul(key_message.as_ref())?
-        .to_str()?
-        .to_string();
+    let key_message_result = CStr::from_bytes_with_nul(key_message.as_ref());
 
-    Ok(key_message)
+    let key_message_cstr = if let Ok(msg) = key_message_result {
+        msg
+    } else {
+        key_message.push(0);
+        CStr::from_bytes_with_nul(key_message.as_ref())?
+    };
+
+    let key_message_string = key_message_cstr.to_str()?.to_string();
+
+    Ok(key_message_string)
+}
+
+pub unsafe fn codes_get_bytes(handle: *mut codes_handle, key: &str) -> Result<Vec<u8>, CodesError> {
+    let mut key_size = codes_get_length(handle, key)?;
+    let key = CString::new(key).unwrap();
+
+    let mut buffer: Vec<u8> = vec![0; key_size as usize];
+
+    let error_code = eccodes_sys::codes_get_bytes(
+        handle,
+        key.as_ptr(),
+        buffer.as_mut_ptr().cast::<u8>(),
+        &mut key_size as *mut u64,
+    );
+
+    if error_code != 0 {
+        let err: CodesInternal = FromPrimitive::from_i32(error_code).unwrap();
+        return Err(err.into());
+    }
+
+    Ok(buffer)
 }
 
 pub unsafe fn codes_get_message_size(handle: *mut codes_handle) -> Result<u64, CodesError> {
@@ -223,7 +261,9 @@ pub unsafe fn codes_get_message_size(handle: *mut codes_handle) -> Result<u64, C
     Ok(size)
 }
 
-pub unsafe fn codes_get_message(handle: *mut codes_handle) -> Result<(*const c_void, u64), CodesError> {
+pub unsafe fn codes_get_message(
+    handle: *mut codes_handle,
+) -> Result<(*const c_void, u64), CodesError> {
     let buffer_size = codes_get_message_size(handle)?;
 
     let buffer: Vec<u8> = vec![0; buffer_size as usize];
@@ -258,11 +298,7 @@ pub unsafe fn codes_handle_new_from_message(
 ) -> *mut codes_handle {
     let default_context: *mut codes_context = ptr::null_mut();
 
-    eccodes_sys::codes_handle_new_from_message(
-        default_context,
-        message_buffer_ptr,
-        message_size,
-    )
+    eccodes_sys::codes_handle_new_from_message(default_context, message_buffer_ptr, message_size)
 }
 
 pub unsafe fn codes_get_message_copy(handle: *mut codes_handle) -> Result<Vec<u8>, CodesError> {
@@ -270,7 +306,7 @@ pub unsafe fn codes_get_message_copy(handle: *mut codes_handle) -> Result<Vec<u8
 
     let mut buffer: Vec<u8> = vec![0; buffer_size as usize];
 
-    let mut message_size: u64 = 0;
+    let mut message_size: u64 = buffer_size;
 
     let error_code = eccodes_sys::codes_get_message_copy(
         handle,
@@ -302,3 +338,120 @@ pub unsafe fn codes_handle_new_from_message_copy(message_buffer: &[u8]) -> *mut
         message_buffer.len() as u64,
     )
 }
+
+pub unsafe fn codes_keys_iterator_new(
+    handle: *mut codes_handle,
+    flags: u32,
+    namespace: &str,
+) -> *mut codes_keys_iterator {
+    let namespace = CString::new(namespace).unwrap();
+
+    eccodes_sys::codes_keys_iterator_new(handle, u64::from(flags), namespace.as_ptr())
+}
+
+pub unsafe fn codes_keys_iterator_delete(
+    keys_iterator: *mut codes_keys_iterator,
+) -> Result<(), CodesError> {
+    let error_code = eccodes_sys::codes_keys_iterator_delete(keys_iterator);
+
+    if error_code != 0 {
+        let err: CodesInternal = FromPrimitive::from_i32(error_code).unwrap();
+        return Err(err.into());
+    }
+
+    Ok(())
+}
+
+pub unsafe fn codes_keys_iterator_next(keys_iterator: *mut codes_keys_iterator) -> bool {
+    let next_item_exists = eccodes_sys::codes_keys_iterator_next(keys_iterator);
+
+    next_item_exists == 1
+}
+
+pub unsafe fn codes_keys_iterator_get_name(
+    keys_iterator: *mut codes_keys_iterator,
+) -> Result<String, CodesError> {
+    let name_pointer = eccodes_sys::codes_keys_iterator_get_name(keys_iterator);
+
+    let name_c_str = CStr::from_ptr(name_pointer);
+    let name_str = name_c_str.to_str()?;
+    let name_string = name_str.to_owned();
+
+    Ok(name_string)
+}
+
+pub unsafe fn codes_grib_nearest_new(
+    handle: *mut codes_handle,
+) -> Result<*mut codes_nearest, CodesError> {
+    let mut error_code: i32 = 0;
+
+    let nearest = eccodes_sys::codes_grib_nearest_new(handle, &mut error_code);
+
+    if error_code != 0 {
+        let err: CodesInternal = FromPrimitive::from_i32(error_code).unwrap();
+        return Err(err.into());
+    }
+
+    Ok(nearest)
+}
+
+pub unsafe fn codes_grib_nearest_delete(nearest: *mut codes_nearest) -> Result<(), CodesError> {
+    let error_code = eccodes_sys::codes_grib_nearest_delete(nearest);
+
+    if error_code != 0 {
+        let err: CodesInternal = FromPrimitive::from_i32(error_code).unwrap();
+        return Err(err.into());
+    }
+
+    Ok(())
+}
+
+pub unsafe fn codes_grib_nearest_find(
+    handle: *mut codes_handle,
+    nearest: *mut codes_nearest,
+    lat: f64,
+    lon: f64,
+) -> Result<[NearestGridpoint; 4], CodesError> {
+    // such flags are set because find nearest for given nearest is always
+    // called on the same grib message
+    let flags = CODES_NEAREST_SAME_GRID + CODES_NEAREST_SAME_DATA;
+
+    let mut output_lats = [0_f64; 4];
+    let mut output_lons = [0_f64; 4];
+    let mut output_values = [0_f64; 4];
+    let mut output_distances = [0_f64; 4];
+    let mut output_indexes = [0_i32; 4];
+
+    let mut length: u64 = 4;
+
+    let error_code = eccodes_sys::codes_grib_nearest_find(
+        nearest,
+        handle,
+        lat,
+        lon,
+        u64::from(flags),
+        &mut output_lats as *mut f64,
+        &mut output_lons as *mut f64,
+        &mut output_values as *mut f64,
+        &mut output_distances as *mut f64,
+        &mut output_indexes as *mut i32,
+        &mut length as *mut u64,
+    );
+
+    if error_code != 0 {
+        let err: CodesInternal = FromPrimitive::from_i32(error_code).unwrap();
+        return Err(err.into());
+    }
+
+    let mut output = [NearestGridpoint::default(); 4];
+
+    for i in 0..4 {
+        output[i].lat = output_lats[i];
+        output[i].lon = output_lons[i];
+        output[i].distance = output_distances[i];
+        output[i].index = output_indexes[i];
+        output[i].value = output_values[i];
+    }
+
+    Ok(output)
+}
